import sqlparse
from sql_metadata import Parser
from sqlinks.app.Objects import Collection


def parse_statement(parser, collection: Collection, statement, statement_type: str):
    statement = str(statement)

    if statement_type == "CTAS":
        l = statement.split(" ")
        if l[0] == "create" and l[1] == "table" and l[3] == "as":
            target_table_sid = l[2]

    elif statement_type == "SELECT_INTO":
        if "into" in statement:
            l = [l for l in statement.split(" ") if l != ""]
            target_table_sid = l[l.index("into") + 1]

    # TODO: Fix the below
    # Currently just skipping any instances where no tables are picked up.
    # They should be picked up, but aren't
    try:
        tables = parser.tables
    except:
        print("Failed to load parser.tables. Skipping statement.")
        return collection

    # Create a table object for every table sid that's found
    # Create a link object for every table sid that's found (with the exception of the target_table_sid)
    for table_sid in parser.tables:
        collection.add_table(sid=table_sid)

        if table_sid != target_table_sid:
            collection.add_link(
                source_table_sid=table_sid, target_table_sid=target_table_sid
            )

    return collection


def clean(raw):
    """
    Clean the statement, returning a list of individual words.

    Args:
        statement (_type_): _description_

    Returns:
        _type_: _description_
    """
    # Remove new lines to negate different formats
    clean = str(raw).lower()
    clean = clean.replace("\n", " ")
    clean = clean.replace("\t", " ")

    # statement = statement.split(' ')
    # statement = [s for s in statement if s != '']
    return clean


def add_to_collection(collection, statements):
    # Looping through all SQL statements in the file, determine what type of statement it is. Add to mapping file accordingly.
    for statement in statements:
        statement_type = statement.get_type()

        if statement_type in ["UNKNOWN", "DROP"]:
            continue
        else:
            parser = Parser(str(statement))

        # Ignore temporary tables
        # TODO: Create functionality to parse temporary tables
        if "create temporary table" in str(statement):
            continue

        # Ignore unpivot statements
        # TODO: Create functionality to parse unpivots
        elif "unpivot" in str(statement):
            continue

        # Parse CTAS statements
        elif statement_type == "CREATE" and "as select" in str(statement):
            collection = parse_statement(
                parser=parser,
                collection=collection,
                statement=statement,
                statement_type="CTAS",
            )

        # Parse SELECT INTO statements
        elif statement_type == "SELECT" and "into" in str(statement):
            collection = parse_statement(
                parser=parser,
                collection=collection,
                statement=statement,
                statement_type="SELECT_INTO",
            )

    return collection


def main(collection: Collection, path):
    try:
        with open(path, "r") as f:
            raw = f.read()
    except:
        print(f"WARNING: Encoding failure {path}")
        return collection

    statements = sqlparse.parse(clean(raw))

    collection = add_to_collection(collection=collection, statements=statements)

    return collection



import argparse
import logging
import os
from pathlib import Path

from sqlinks.app.Objects import Collection

from sqlinks.app import parse as parse
from sqlinks.app import draw as draw
from sqlinks.app import scan as scan
from sqlinks.app import open as open


def get_args():
    """_summary_

    Returns:
        _type_: _description_
    """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-p", "--path", default=".", help="Path to scan across | Default value = ."
    )

    parser.add_argument(
        "-o",
        "--output",
        default=".",
        help="Path to output diagrams | Set to 'REL' to save adjacent to scripts | Default value = .",
    )

    parser.add_argument(
        "-l",
        "--logging_level",
        default="error",
        help="Set the level of logging messages to receive | Options are [debug, info, warning, error, critical] | Default value = debug",
    )

    parser.add_argument(
        "-au",
        "--auto_open",
        default="n",
        help="Set whether the output file should be auto-opened or not after execution | Options are [y, n] | Default value = n",
    )

    args = parser.parse_args()
    return args


def set_logger(logging_level: str):
    """
    Set the logging level based off passed arguments
    """

    if logging_level == "debug":
        level = logging.DEBUG
    elif logging_level == "info":
        level = logging.INFO
    elif logging_level == "warning":
        level = logging.WARNING
    elif logging_level == "error":
        level = logging.ERROR
    elif logging_level == "critical":
        level = logging.CRITICAL
    else:
        level = logging.DEBUG

    logging.basicConfig(level=level)


def setup(working_dir: Path):
    """
    Set up working environment
    """
    os.makedirs(working_dir, exist_ok=True)


def cleanup(working_dir: Path):
    """
    Cleanup working environment
    """
    filelist = [f for f in os.listdir(working_dir)]
    for f in filelist:
        os.remove(os.path.join(working_dir, f))
    os.rmdir(working_dir)


def print_args(args):
    """Print args

    Args:
        args ():
    """
    print("\n")
    print("------------------")
    print("Selected settings")
    print("------------------")

    for key, value in vars(args).items():
        print(f"{key}: {value}")

    print("\n")
    return None


if __name__ == "__main__":
    args = get_args()
    print_args(args)
    set_logger(args.logging_level)

    working_dir = Path("./working-files/")
    output_filename = Path(args.output + "/output.md")

    setup(working_dir=working_dir)

    paths = scan.main(args.path)

    # Instantiate empty ObjectCollection
    collection = Collection()

    for i in range(0, len(paths)):
        path = paths[i]
        print("\n")
        print(f"-" * 10)
        print(f"[{i+1}]/[{len(paths)}]")
        print(f"-" * 10)
        print(f"File: {path['relative_path']}")
        collection = parse.main(collection=collection, path=path["absolute_path"])

    print("\nGenerating diagram...")

    draw.main(output_filename=output_filename, collection=collection)

    cleanup(working_dir=working_dir)

    # Open the output file if auto_open is selected
    if args.auto_open == "y":
        open.open_drawing(output_filename)



import os
from pathlib import Path


def main(path):
    """
    For the provided path, return list of all SQL files
    """
    paths = []

    for dirpath, dirnames, files in os.walk(path):
        for name in files:
            if name.lower().endswith("sql"):
                relative_path = os.path.join(dirpath, name)
                absolute_path = os.path.abspath(relative_path)
                paths.append(
                    {
                        "absolute_path": Path(absolute_path),
                        "relative_path": Path(relative_path),
                    }
                )

    return paths



import os
import logging


def open_drawing(output_filename):
    """_summary_
    Args:
        output_filename (_type_): _description_
    """
    try:
        os.system("open {}".format(output_filename))
    except:
        logging.error("Unable to open .drawio file")



"""draw.py."""
from jinja2 import Environment, FileSystemLoader
import logging
from pathlib import Path


def gen_drawing(collection):
    """_summary_
    Args:
        tables (_type_): _description_

    Returns:
        _type_: _description_
    """
    file_loader = FileSystemLoader(Path(__file__).parent / "templates/")
    env = Environment(loader=file_loader)

    template = env.get_template("mermaid_template.md")

    output = template.render(collection=collection)
    return output


def save_drawing(output, output_filename):
    """_summary_
    Args:
        output (_type_): _description_
        output_filename (_type_): _description_
    """
    with open(output_filename, "w+") as f:
        f.write(output)


def save_drawing(output, output_filename):
    """_summary_

    Args:
        output (_type_): _description_
        output_filename (_type_): _description_
    """
    with open(output_filename, "w+") as f:
        f.write(output)


def main(collection, output_filename: Path):
    """_summary_

    Args:
        working_dir (Path): _description_
        output_filename (Path): _description_
    """

    output = gen_drawing(collection=collection)
    logging.debug(output_filename)

    save_drawing(output, output_filename)



from setuptools import setup, find_packages

VERSION = "0.0.3"
DESCRIPTION = "Diagramming SQL data pipelines."
LONG_DESCRIPTION = "Documenting data pipelines is hard. Not only is it complex, but it's also time consuming. When working with big teams it becomes near impossible; documentation is out of date as soon as it's completed as there's continual change. **sqlinks** hopes to remedy this by allowing you to generate flow diagrams programatically. This allows for reproducable and clear diagrams that, by using [diagrams.net](diagrams.net), are also easily accessible by non-technical users."

# Setting up
setup(
    name="sqlinks",
    version=VERSION,
    author="Dominic Herriott",
    author_email="dominicherriott@outlook.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=["sqlparse", "sql_metadata", "jinja2"],
    keywords=["python", "documentation", "sql"],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    package_data={"sqlinks": ["app/templates/*.xml"]},
)



def deconstruct_sid(sid, how):
    if how == "column":
        schema, table = sid.split(".")[:2]
        schema_sid = schema
        table_sid = ".".join([schema, table])
        column_sid = sid
        return schema_sid, table_sid, column_sid

    elif how == "table":
        # For handling [schema.table] table references
        if sid.count(".") == 1:
            schema, table = sid.split(".")
            schema_sid = schema
            table_sid = ".".join([schema, table])
            return schema_sid, table_sid

        # For hanlding [db.schema.table] table references
        elif sid.count(".") == 2:
            db, schema, table = sid.split(".")
            schema_sid = ".".join([db, schema])
            table_sid = sid
            return schema_sid, table_sid

    else:
        raise Exception("'how' only takes values of 'column' or 'table")


def clean_sid(sid):
    sid = sid.replace("[", "")
    sid = sid.replace("]", "")
    sid = sid.replace("'", "")
    sid = sid.replace('"', "")
    return sid



from sqlinks.app.helpers import deconstruct_sid, clean_sid


class Object:
    object_counter = 1

    def __init__(self, sid):
        # String ID
        self.sid = sid
        # Numerical ID
        self.nid = Object.object_counter
        Object.object_counter += 1


class Table(Object):
    def __init__(self, sid: str):
        Object.__init__(self, sid)

        schema_sid, table_sid = deconstruct_sid(sid, how="table")

        self.sid = table_sid
        self.schema_sid = schema_sid
        self.name = table_sid.split(".")[1]


class Link(Object):
    def __init__(self, source_table_sid: str, target_table_sid: str):
        Object.__init__(self, sid=None)
        self.source_table_sid = source_table_sid
        self.target_table_sid = target_table_sid


class Collection:
    def __init__(self):
        self.tables = {}
        self.links = []

    def add_table(self, sid: str):
        sid = clean_sid(sid)

        if sid in self.tables:
            return
        self.tables[sid] = Table(sid=sid)

    def add_link(self, source_table_sid: int, target_table_sid: int):
        source_table_sid = clean_sid(source_table_sid)
        target_table_sid = clean_sid(target_table_sid)

        self.links.append(
            Link(source_table_sid=source_table_sid, target_table_sid=target_table_sid)
        )
